@@ -1,24 +1,53 @@
 import copy
+import functools
+import importlib.resources
 import itertools
 import math
 import re
 import sys
+from abc import ABC, abstractmethod
 from dataclasses import dataclass
 from functools import lru_cache
-from typing import Set, Generator, Tuple, List, Dict, Union, TypeVar, Sequence, cast, Callable, Iterable, Any, Optional
+from typing import (
+    Set,
+    Generator,
+    Tuple,
+    List,
+    Dict,
+    Union,
+    TypeVar,
+    Sequence,
+    cast,
+    Callable,
+    Iterable,
+    Any,
+    Optional,
+    Generic,
+    Iterator,
+)
 
-from isla.type_defs import Path, Grammar, ParseTree, ImmutableGrammar, CanonicalGrammar, ImmutableList
+from isla.type_defs import (
+    Path,
+    Grammar,
+    ParseTree,
+    ImmutableGrammar,
+    CanonicalGrammar,
+    ImmutableList,
+)
 
-S = TypeVar('S')
-T = TypeVar('T')
+R = TypeVar("R")
+S = TypeVar("S")
+T = TypeVar("T")
 
 
 def is_path(maybe_path: Any) -> bool:
-    return isinstance(maybe_path, tuple) and all(isinstance(elem, int) for elem in maybe_path)
+    return isinstance(maybe_path, tuple) and all(
+        isinstance(elem, int) for elem in maybe_path
+    )
 
 
-def pop(l: List[S], default: T = None, index=0) -> Union[S, T]:
-    return default if not l else l.pop(index)
+def pop(a_list: List[S], default: T = None, index=0) -> Union[S, T]:
+    return default if not a_list else a_list.pop(index)
 
 
 def replace_line_breaks(inp: str) -> str:
@@ -26,14 +55,16 @@ def replace_line_breaks(inp: str) -> str:
 
 
 def parent_reflexive(path_1: Path, path_2: Path) -> bool:
-    return len(path_1) <= len(path_2) and path_1 == path_2[:len(path_1)]
+    return len(path_1) <= len(path_2) and path_1 == path_2[: len(path_1)]
 
 
 def parent_or_child(path_1: Path, path_2: Path) -> bool:
-    return path_1[:len(path_2)] == path_2 or path_2[:len(path_1)] == path_1
+    return path_1[: len(path_2)] == path_2 or path_2[: len(path_1)] == path_1
 
 
-def path_iterator(tree: ParseTree, path: Path = ()) -> Generator[Tuple[Path, ParseTree], None, None]:
+def path_iterator(
+    tree: ParseTree, path: Path = ()
+) -> Generator[Tuple[Path, ParseTree], None, None]:
     yield path, tree
     if tree[1] is not None:
         for i, child in enumerate(tree[1]):
@@ -45,7 +76,9 @@ def delete_unreachable(grammar: Grammar) -> None:
         del grammar[unreachable]
 
 
-def replace_tree_path(in_tree: ParseTree, path: Path, replacement_tree: ParseTree) -> ParseTree:
+def replace_tree_path(
+    in_tree: ParseTree, path: Path, replacement_tree: ParseTree
+) -> ParseTree:
     """Returns tree where replacement_tree has been inserted at `path` instead of the original subtree"""
     node, children = in_tree
 
@@ -53,9 +86,11 @@ def replace_tree_path(in_tree: ParseTree, path: Path, replacement_tree: ParseTre
         return replacement_tree
 
     head = path[0]
-    new_children = (children[:head] +
-                    [replace_tree_path(children[head], path[1:], replacement_tree)] +
-                    children[head + 1:])
+    new_children = (
+        children[:head]
+        + [replace_tree_path(children[head], path[1:], replacement_tree)]
+        + children[head + 1 :]
+    )
 
     return node, new_children
 
@@ -81,11 +116,12 @@ TRAVERSE_POSTORDER = 1
 
 
 def traverse(
-        tree: ParseTree,
-        action: Callable[[Path, ParseTree], None],
-        abort_condition: Callable[[Path, ParseTree], bool] = lambda p, n: False,
-        kind: int = TRAVERSE_PREORDER,
-        reverse: bool = False) -> None:
+    tree: ParseTree,
+    action: Callable[[Path, ParseTree], None],
+    abort_condition: Callable[[Path, ParseTree], bool] = lambda p, n: False,
+    kind: int = TRAVERSE_PREORDER,
+    reverse: bool = False,
+) -> None:
     stack_1: List[Tuple[Path, ParseTree]] = [((), tree)]
     stack_2: List[Tuple[Path, ParseTree]] = []
 
@@ -125,13 +161,13 @@ def tree_to_string(tree: ParseTree) -> str:
         symbol, children = stack.pop(0)
 
         if not children:
-            result.append('' if is_nonterminal(symbol) else symbol)
+            result.append("" if is_nonterminal(symbol) else symbol)
             continue
 
         for child in reversed(children):
             stack.insert(0, child)
 
-    return ''.join(result)
+    return "".join(result)
 
 
 def tree_depth(tree: ParseTree) -> int:
@@ -155,10 +191,7 @@ def canonical(grammar: Grammar) -> CanonicalGrammar:
         if isinstance(expansion, tuple):
             expansion = expansion[0]
 
-        return [
-            token
-            for token in RE_NONTERMINAL.split(expansion)
-            if token]
+        return [token for token in RE_NONTERMINAL.split(expansion) if token]
 
     return {
         k: [split(expression) for expression in alternatives]
@@ -166,7 +199,9 @@ def canonical(grammar: Grammar) -> CanonicalGrammar:
     }
 
 
-def dict_of_lists_to_list_of_dicts(dict_of_lists: Dict[S, Iterable[T]]) -> List[Dict[S, T]]:
+def dict_of_lists_to_list_of_dicts(
+    dict_of_lists: Dict[S, Iterable[T]]
+) -> List[Dict[S, T]]:
     keys = list(dict_of_lists.keys())
     list_of_values = [dict_of_lists[key] for key in keys]
     product = list(itertools.product(*list_of_values))
@@ -182,32 +217,34 @@ def powerset(iterable):
     """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
     s = list(iterable)
     return itertools.chain.from_iterable(
-        itertools.combinations(s, r)
-        for r in range(len(s) + 1))
+        itertools.combinations(s, r) for r in range(len(s) + 1)
+    )
 
 
 def weighted_geometric_mean(seq: Sequence[float], weights: Sequence[float]) -> float:
     assert len(seq) == len(weights)
     assert all(w >= 0 for w in weights)
 
-    return (math.prod([(n + 1) ** w for n, w in zip(seq, weights) if w > 0]) **
-            (1 / sum([w for w in weights if w > 0]))) - 1
+    return (
+        math.prod([(n + 1) ** w for n, w in zip(seq, weights) if w > 0])
+        ** (1 / sum([w for w in weights if w > 0]))
+    ) - 1
 
 
 def grammar_to_immutable(grammar: Grammar) -> ImmutableGrammar:
-    return cast(ImmutableGrammar, tuple({k: tuple(v) for k, v in grammar.items()}.items()))
+    return cast(
+        ImmutableGrammar, tuple({k: tuple(v) for k, v in grammar.items()}.items())
+    )
 
 
 def grammar_to_mutable(grammar: ImmutableGrammar) -> Grammar:
     return {nonterminal: list(expansion) for nonterminal, expansion in grammar}
 
 
-def immutable_to_grammar(immutable: ImmutableGrammar) -> Grammar:
-    return {k: list(v) for k, v in dict(immutable).items()}
-
-
-def nested_list_to_tuple(l: List[Union[T, List[T]]]) -> Tuple[Union[T, Tuple[T, ...]], ...]:
-    return tuple([tuple(elem) if isinstance(elem, list) else elem for elem in l])
+def nested_list_to_tuple(
+    a_list: List[Union[T, List[T]]]
+) -> Tuple[Union[T, Tuple[T, ...]], ...]:
+    return tuple([tuple(elem) if isinstance(elem, list) else elem for elem in a_list])
 
 
 def assertions_activated() -> bool:
@@ -223,16 +260,17 @@ def assertions_activated() -> bool:
 
 
 def split_str_with_nonterminals(expression: str) -> List[str]:
-    return [token for token in re.split(
-        RE_NONTERMINAL, expression) if token]
+    return [token for token in re.split(RE_NONTERMINAL, expression) if token]
 
 
-def cluster_by_common_elements(l: Sequence[T], f: Callable[[T], Set[S]]) -> List[List[T]]:
+def cluster_by_common_elements(
+    a_list: Sequence[T], f: Callable[[T], Set[S]]
+) -> List[List[T]]:
     """
     Clusters elements of l by shared elements. Elements of interest are obtained using f.
     For instance, to cluster a list of lists based on common list elements:
 
-    >>> cluster_by_common_elements([[1,2], [2,3], [3,4], [5,6]], lambda l: {e for e in l})
+    >>> cluster_by_common_elements([[1,2], [2,3], [3,4], [5,6]], lambda l: {e for e in a_list})
 
     Gives [[[2, 3], [3, 4], [1, 2]], [[5, 6]]].
 
@@ -240,16 +278,20 @@ def cluster_by_common_elements(l: Sequence[T], f: Callable[[T], Set[S]]) -> List
 
     Outputs [[2, 5], [3, 6], [4, 7, 1]].
 
-    :param l: The list to cluster.
+    :param a_list: The list to cluster.
     :param f: The function to determine commonality.
     :return: The clusters w.r.t. f.
     """
-    clusters = [[e2 for e2 in l if not f(e1).isdisjoint(f(e2))] for e1 in l]
+    clusters = [[e2 for e2 in a_list if not f(e1).isdisjoint(f(e2))] for e1 in a_list]
 
     result = []
     for c in clusters:
         # Merge clusters with common elements...
-        clusters_with_common_elements = [c1 for c1 in result if any(not f(e).isdisjoint(f(e1)) for e in c for e1 in c1)]
+        clusters_with_common_elements = [
+            c1
+            for c1 in result
+            if any(not f(e).isdisjoint(f(e1)) for e in c for e1 in c1)
+        ]
         for c1 in clusters_with_common_elements:
             result.remove(c1)
 
@@ -276,7 +318,7 @@ def crange(character_start: str, character_end: str) -> List[str]:
     return [chr(i) for i in range(ord(character_start), ord(character_end) + 1)]
 
 
-RE_EXTENDED_NONTERMINAL = re.compile(r'(<[^<> ]*>[?+*])')
+RE_EXTENDED_NONTERMINAL = re.compile(r"(<[^<> ]*>[?+*])")
 
 
 def extended_nonterminals(expansion: str) -> List[str]:
@@ -321,8 +363,9 @@ def convert_ebnf_operators(ebnf_grammar: Grammar) -> Grammar:
             for extended_symbol in extended_symbols:
                 operator = extended_symbol[-1:]
                 original_symbol = extended_symbol[:-1]
-                assert original_symbol in ebnf_grammar, \
-                    f"{original_symbol} is not defined in grammar"
+                assert (
+                    original_symbol in ebnf_grammar
+                ), f"{original_symbol} is not defined in grammar"
 
                 new_sym = new_symbol(grammar, original_symbol)
 
@@ -338,18 +381,17 @@ def convert_ebnf_operators(ebnf_grammar: Grammar) -> Grammar:
                 else:
                     grammar[nonterminal][i] = new_exp
 
-                if operator == '?':
+                if operator == "?":
                     grammar[new_sym] = ["", original_symbol]
-                elif operator == '*':
+                elif operator == "*":
                     grammar[new_sym] = ["", original_symbol + new_sym]
-                elif operator == '+':
-                    grammar[new_sym] = [
-                        original_symbol, original_symbol + new_sym]
+                elif operator == "+":
+                    grammar[new_sym] = [original_symbol, original_symbol + new_sym]
 
     return grammar
 
 
-RE_PARENTHESIZED_EXPR = re.compile(r'\([^()]*\)[?+*]')
+RE_PARENTHESIZED_EXPR = re.compile(r"\([^()]*\)[?+*]")
 
 
 def parenthesized_expressions(expansion: str) -> List[str]:
@@ -404,7 +446,7 @@ def convert_ebnf_grammar(ebnf_grammar: Grammar) -> Grammar:
     return convert_ebnf_operators(convert_ebnf_parentheses(ebnf_grammar))
 
 
-RE_NONTERMINAL = re.compile(r'(<[^<> ]*>)')
+RE_NONTERMINAL = re.compile(r"(<[^<> ]*>)")
 
 
 @lru_cache(maxsize=None)
@@ -431,9 +473,8 @@ def transitive_closure(relation: Iterable[Tuple[S, T]]) -> Set[Tuple[S, T]]:
     closure = set(relation)
     while True:
         new_relations: Set[Tuple[S, T]] = {
-            (x, w)
-            for x, y in closure
-            for q, w in closure if q == y}
+            (x, w) for x, y in closure for q, w in closure if q == y
+        }
 
         closure_until_now = closure | new_relations
 
@@ -446,11 +487,12 @@ def transitive_closure(relation: Iterable[Tuple[S, T]]) -> Set[Tuple[S, T]]:
 
 
 def start_symbol():
-    return '<start>'
+    return "<start>"
 
 
-def def_used_nonterminals(grammar: Grammar, _start_symbol: str = start_symbol()) -> \
-        Tuple[Optional[Set[str]], Optional[Set[str]]]:
+def def_used_nonterminals(
+    grammar: Grammar, _start_symbol: str = start_symbol()
+) -> Tuple[Optional[Set[str]], Optional[Set[str]]]:
     """Return a pair (`defined_nonterminals`, `used_nonterminals`) in `grammar`.
     In case of error, return (`None`, `None`)."""
 
@@ -461,22 +503,26 @@ def def_used_nonterminals(grammar: Grammar, _start_symbol: str = start_symbol())
         defined_nonterminals.add(defined_nonterminal)
         expansions = grammar[defined_nonterminal]
         if not isinstance(expansions, list):
-            print(repr(defined_nonterminal) + ": expansion is not a list",
-                  file=sys.stderr)
+            print(
+                repr(defined_nonterminal) + ": expansion is not a list", file=sys.stderr
+            )
             return None, None
 
         if len(expansions) == 0:
-            print(repr(defined_nonterminal) + ": expansion list empty",
-                  file=sys.stderr)
+            print(repr(defined_nonterminal) + ": expansion list empty", file=sys.stderr)
             return None, None
 
         for expansion in expansions:
             if isinstance(expansion, tuple):
                 expansion = expansion[0]
             if not isinstance(expansion, str):
-                print(repr(defined_nonterminal) + ": "
-                      + repr(expansion) + ": not a string",
-                      file=sys.stderr)
+                print(
+                    repr(defined_nonterminal)
+                    + ": "
+                    + repr(expansion)
+                    + ": not a string",
+                    file=sys.stderr,
+                )
                 return None, None
 
             for used_nonterminal in nonterminals(expansion):
@@ -485,7 +531,9 @@ def def_used_nonterminals(grammar: Grammar, _start_symbol: str = start_symbol())
     return defined_nonterminals, used_nonterminals
 
 
-def reachable_nonterminals(grammar: Grammar, _start_symbol: str = start_symbol()) -> Set[str]:
+def reachable_nonterminals(
+    grammar: Grammar, _start_symbol: str = start_symbol()
+) -> Set[str]:
     reachable = set()
 
     def _find_reachable_nonterminals(grammar, symbol):
@@ -500,17 +548,23 @@ def reachable_nonterminals(grammar: Grammar, _start_symbol: str = start_symbol()
     return reachable
 
 
-def unreachable_nonterminals(grammar: Grammar, _start_symbol=start_symbol()) -> Set[str]:
+def unreachable_nonterminals(
+    grammar: Grammar, _start_symbol=start_symbol()
+) -> Set[str]:
     return grammar.keys() - reachable_nonterminals(grammar, _start_symbol)
 
 
-def is_valid_grammar(grammar: Grammar, _start_symbol: str = start_symbol(), ) -> bool:
+def is_valid_grammar(
+    grammar: Grammar,
+    _start_symbol: str = start_symbol(),
+) -> bool:
     """Check if the given `grammar` is valid.
-       `start_symbol`: optional start symbol (default: `<start>`)
-       `supported_opts`: options supported (default: none)"""
+    `start_symbol`: optional start symbol (default: `<start>`)
+    `supported_opts`: options supported (default: none)"""
 
-    defined_nonterminals, used_nonterminals = \
-        def_used_nonterminals(grammar, _start_symbol)
+    defined_nonterminals, used_nonterminals = def_used_nonterminals(
+        grammar, _start_symbol
+    )
     if defined_nonterminals is None or used_nonterminals is None:
         return False
 
@@ -520,11 +574,9 @@ def is_valid_grammar(grammar: Grammar, _start_symbol: str = start_symbol(), ) ->
         used_nonterminals.add(start_symbol())
 
     for unused_nonterminal in defined_nonterminals - used_nonterminals:
-        print(repr(unused_nonterminal) + ": defined, but not used",
-              file=sys.stderr)
+        print(repr(unused_nonterminal) + ": defined, but not used", file=sys.stderr)
     for undefined_nonterminal in used_nonterminals - defined_nonterminals:
-        print(repr(undefined_nonterminal) + ": used, but not defined",
-              file=sys.stderr)
+        print(repr(undefined_nonterminal) + ": used, but not defined", file=sys.stderr)
 
     # Symbols must be reachable either from <start> or given start symbol
     unreachable = unreachable_nonterminals(grammar, _start_symbol)
@@ -536,8 +588,10 @@ def is_valid_grammar(grammar: Grammar, _start_symbol: str = start_symbol(), ) ->
             msg_start_symbol += " or " + start_symbol()
 
     for unreachable_nonterminal in unreachable:
-        print(repr(unreachable_nonterminal) + ": unreachable from " + msg_start_symbol,
-              file=sys.stderr)
+        print(
+            repr(unreachable_nonterminal) + ": unreachable from " + msg_start_symbol,
+            file=sys.stderr,
+        )
 
     return used_nonterminals == defined_nonterminals and len(unreachable) == 0
 
@@ -559,11 +613,13 @@ class lazystr:
         return str(self.c())
 
 
-def replace_in_list(l: List[T], repl: S | List[S], idx: int) -> List[T | S]:
-    assert -len(l) <= idx < len(l)
+def replace_in_list(a_list: List[T], repl: S | List[S], idx: int) -> List[T | S]:
+    assert -len(a_list) <= idx < len(a_list)
     if idx < 0:
-        idx = len(l) + idx
-    return l[0:idx] + (repl if isinstance(repl, list) else [repl]) + l[idx + 1:]
+        idx = len(a_list) + idx
+    return (
+        a_list[0:idx] + (repl if isinstance(repl, list) else [repl]) + a_list[idx + 1 :]
+    )
 
 
 def nth_occ(haystack: Sequence[T], needle: T, n: int) -> Optional[int]:
@@ -579,4 +635,117 @@ def nth_occ(haystack: Sequence[T], needle: T, n: int) -> Optional[int]:
 
 
 def list_set(ilist: ImmutableList[T], repl_idx: int, new_elem: T) -> ImmutableList[T]:
-    return tuple([new_elem if idx == repl_idx else elem for idx, elem in enumerate(ilist)])
+    return tuple(
+        [new_elem if idx == repl_idx else elem for idx, elem in enumerate(ilist)]
+    )
+
+
+@dataclass(frozen=True)
+class Monad(ABC, Generic[T]):
+    a: T
+
+    @abstractmethod
+    def bind(self, f: Callable[[T], "Monad[S]"]) -> "Monad[S]":
+        raise NotImplementedError()
+
+
+@dataclass(frozen=True)
+class MonadPlus(Generic[T], Monad[T]):
+    @staticmethod
+    @abstractmethod
+    def nothing() -> "MonadPlus[T]":
+        raise NotImplementedError()
+
+    @abstractmethod
+    def mplus(self, other: "MonadPlus[T]") -> "MonadPlus[T]":
+        raise NotImplementedError()
+
+    @abstractmethod
+    def lazy_mplus(self, f: Callable[[S], "MonadPlus[T]"], arg: S) -> "MonadPlus[T]":
+        raise NotImplementedError()
+
+
+@dataclass(frozen=True)
+class Maybe(Generic[T], MonadPlus[Optional[T]]):
+    a: Optional[T]
+
+    def bind(self, f: Callable[[T], "Maybe[S]"]) -> "Maybe[S]":
+        return self if self.a is None else f(self.a)
+
+    def map(self, f: Callable[[T], S]) -> "Maybe[S]":
+        return self if self.a is None else Maybe(f(self.a))
+
+    def orelse(self, f: Callable[[], S]) -> "Maybe[S]":
+        assert callable(f)
+        return self if self.a is not None else Maybe(f())
+
+    @staticmethod
+    def nothing() -> "Maybe[T]":
+        return Maybe(None)
+
+    @staticmethod
+    def from_iterator(iterator: Iterator[T]) -> "Maybe[T]":
+        try:
+            return Maybe(next(iterator))
+        except StopIteration:
+            return Maybe.nothing()
+
+    def mplus(self, other: "Maybe[T]") -> "Maybe[T]":
+        return other if self.a is None else self
+
+    def lazy_mplus(self, f: Callable[[S, ...], "Maybe[T]"], *args: S) -> "Maybe[T]":
+        return f(*args) if self.a is None else self
+
+    def if_present(self, f: Callable[[T], None]) -> None:
+        if self.a is not None:
+            f(self.a)
+
+    def is_present(self) -> bool:
+        return self.a is not None
+
+    def raise_if_not_present(self, exc: Callable[[], Exception]) -> "Maybe[T]":
+        if self.a is None:
+            raise exc()
+
+        return self
+
+    def get(self) -> T:
+        if self.a is None:
+            raise AttributeError("No element present")
+        return self.a
+
+    def __add__(
+        self,
+        other: "Maybe[T]" | Tuple[Callable[[S, ...], "Maybe[T]"], S],
+    ) -> "Maybe[T]":
+        if isinstance(other, Maybe):
+            return self.mplus(other)
+
+        assert isinstance(other, tuple)
+        assert callable(other[0])
+        return self.lazy_mplus(*other)
+
+
+def chain_functions(
+    functions: Iterable[Callable[[S, ...], Maybe[T]]], *args: S
+) -> Maybe[T]:
+    return functools.reduce(
+        lambda monad, f: (monad + (f, *args)),
+        functions,
+        Maybe.nothing(),
+    )
+
+
+def is_float(num: Any) -> bool:
+    try:
+        float(num)
+        return True
+    except ValueError:
+        return False
+
+
+def get_isla_resource_file_content(path_to_file: str) -> str:
+    traversable = importlib.resources.files("isla").joinpath(path_to_file)
+    with importlib.resources.as_file(traversable) as path:
+        with open(path, "r") as file:
+            return file.read()
